"""This module is in charge of managing all the trigger services.

Trigger services are the components responsible managing the execution of Augs.

Each trigger service must export the following methods:
- __init__(self, output) - Construct the object, receives a point to the Output object.
- remove_aug(self, aug_id) - Remove an Aug by it's id.
- clear_augs(self) - Remove all Augs.

Services are loaded from the services packages.
"""
from six import itervalues

from .services import BdbLocationService, ImportService, LoggingLocationService, HttpServerService, \
    ExceptionCapturingLocationService

from .exceptions import RookServiceMissing

from .logger import logger


class TriggerServices(object):
    """This class implements all the module's functionality."""

    def __init__(self):
        """Load and initialize all the trigger services."""
        self._services = dict()

    def get_service(self, name):
        """Return the trigger service by it's name."""
        try:
            return self._services[name]
        except KeyError:
            raise RookServiceMissing(name)

    def remove_aug(self, aug_id):
        """Remove an Aug from all services by it's id."""
        for service in itervalues(self._services):
            try:
                service.remove_aug(aug_id)
            except Exception:
                logger.exception("Error when removing aug")

    def clear_augs(self):
        """Remove all Augs from all services."""

        for service in itervalues(self._services):
            service.clear_augs()

    def start(self):
        self._load_services()

    def close(self):
        for service in itervalues(self._services):
            service.close()

        self._services = {}

    def pre_fork(self):
        for service in itervalues(self._services):
            service.pre_fork()

    def post_fork(self):
        for service in itervalues(self._services):
            service.post_fork()

    def _load_services(self):
        try:
            self._services[BdbLocationService.NAME] = BdbLocationService()
            self._services[ImportService.NAME] = ImportService(self._services[BdbLocationService.NAME])
        except:  # lgtm[py/catch-base-exception]
            logger.exception("Error starting location based services")

        try:
            self._services[LoggingLocationService.NAME] = LoggingLocationService()
        except:  # lgtm[py/catch-base-exception]
            logger.exception("Error starting log handler services")

        try:
            self._services[HttpServerService.NAME] = HttpServerService()
        except:  # lgtm[py/catch-base-exception]
            logger.exception("Error starting http server services")

        try:
            self._services[ExceptionCapturingLocationService.NAME] = ExceptionCapturingLocationService()
        except:  # lgtm[py/catch-base-exception]
            logger.exception("Error starting exception capturing handler services")



import sys
import traceback

from rook.exceptions import ToolException
from rook.processor.namespace_serializer import NamespaceSerializer  # lgtm[py/cyclic-import]
from rook.processor.namespaces.python_object_namespace import PythonObjectNamespace
import rook.protobuf.variant_pb2


class Error(object):
    def __init__(self, exc=None, type=None, message=None, parameters=None, traceback=None):
        self.exc = exc

        if isinstance(exc, ToolException):
            self.type = exc.get_type()
            self.message = exc.get_message()
            self.parameters = exc.get_parameters()
        else:
            self.type = type or "Unknown"
            self.message = message
            self.parameters = parameters

        # Extract traceback if was not supplied
        self.traceback = traceback
        if self.traceback is None and sys.exc_info()[2]:
            self.traceback = sys.exc_info()[2]

    def dump(self, error):
        try:
            if self.message:
                error.message = NamespaceSerializer.normalize_string(self.message)
            else:
                error.message = ''
        except Exception: # We are already deep in error handling code - just ignore
            pass

        try:
            error.type = NamespaceSerializer.normalize_string(self.type)
        except Exception: # We are already deep in error handling code - just ignore
            pass

        try:
            serializer = NamespaceSerializer()
            serializer.dump(PythonObjectNamespace(self.parameters), error.parameters, False)
        except Exception: # We are already deep in error handling code - just ignore
            pass

        try:
            serializer = NamespaceSerializer()
            serializer.dump(PythonObjectNamespace(self.exc), error.exc, False)
        except Exception: # We are already deep in error handling code - just ignore
            pass

        try:
            if self.traceback:
                traceback_string = '\n'.join(traceback.format_tb(self.traceback))
                serializer = NamespaceSerializer()
                serializer.dump(PythonObjectNamespace(traceback_string,
                                                      PythonObjectNamespace.ObjectDumpConfig.tailor_limits(traceback_string)),
                                error.traceback, False)
        except Exception: # We are already deep in error handling code - just ignore
            pass

    def dumps(self):
        error = rook.protobuf.variant_pb2.Error()
        self.dump(error)
        return error



import sys
import platform


class LinuxDistribution:
    name = ''
    version = ''

    if platform.system() == "Linux":
        if sys.version_info[0] < 3:
            name, version = platform.linux_distribution()[:2]
        else:
            try:
                import distro
                name = distro.name()
                version = distro.version()
            except ImportError:
                pass

    @staticmethod
    def get_version():
        return LinuxDistribution.version

    @staticmethod
    def get_name():
        return LinuxDistribution.name



import threading


class ExceptionCapturingLocationService(object):
    NAME = "exception_capturing"

    def __init__(self):
        self._lock = threading.RLock()
        self._aug_ids = []

    def get_ids(self):
        return self._aug_ids

    def add_exception_capturing_aug(self, aug):
        with self._lock:
            self._aug_ids.append(aug.aug_id)
            aug.set_active()

    def remove_aug(self, aug_id):
        with self._lock:
            try:
                self._aug_ids.remove(aug_id)
            except ValueError: # if value is not the array then there's nothing to do here
                pass

    def clear_augs(self):
        with self._lock:
            del self._aug_ids[:]

    def close(self):
        self.clear_augs()

    def pre_fork(self):
        if self._lock:
            self._lock.acquire()

    def post_fork(self):
        if self._lock:
            self._lock.release()

    def activation_status(self):
        return len(self._aug_ids) > 0



import sys
import logging
import six
import os
import threading

from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.augs.processor_extensions.namespaces.log_record_namespace import LogRecordNamespace

# For traceback analysis
currentframe = lambda: sys._getframe(3)
_srcfile = os.path.normcase(currentframe.__code__.co_filename)
_logging_srcfile = logging._srcfile

# Copied from logging
CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

_levelNames = {
    CRITICAL : 'CRITICAL',
    ERROR : 'ERROR',
    WARNING : 'WARNING',
    INFO : 'INFO',
    DEBUG : 'DEBUG',
    NOTSET : 'NOTSET',
    'CRITICAL' : CRITICAL,
    'ERROR' : ERROR,
    'WARN' : WARNING,
    'WARNING' : WARNING,
    'INFO' : INFO,
    'DEBUG' : DEBUG,
    'NOTSET' : NOTSET,
}

class LoggingLocationService(object):
    NAME = "logging"

    METHODS_TO_HOOK = ["critical", "error", "warning", "info", "debug"]

    def __init__(self):
        self._lock = threading.RLock()

        self._hooked = False
        self._originalMethods = dict()
        self._originalHandle = None

        self._augs = dict()

    def add_logging_aug(self, aug):
        with self._lock:
            self._installHooks()

            self._augs[aug.aug_id] = aug
            aug.set_active()

    def remove_aug(self, aug_id):
        with self._lock:
            try:
                del self._augs[aug_id]
            except KeyError:
                pass

            if not self._augs:
                self._removeHooks()

    def clear_augs(self):
        with self._lock:
            for aug_id in list(self._augs.keys()):
                self.remove_aug(aug_id)

            self._augs = dict()
            self._removeHooks()

    def close(self):
        self.clear_augs()
        self._removeHooks()

    def pre_fork(self):
        if self._lock:
            self._lock.acquire()

    def post_fork(self):
        if self._lock:
            self._lock.release()

        self._removeHooks()

    def _installHooks(self):
        with self._lock:
            if self._hooked:
                return

            # TODO - hook log
            # TODO - properly set exc_info=1 on exception
            for name in self.METHODS_TO_HOOK:
                self._wrap_logging_method(name)

            self._hooked = True

    # based off https://github.com/getsentry/raven-python/blob/master/raven/breadcrumbs.py
    def _wrap_logging_method(self, name):

        func = logging.Logger.__dict__[name]
        code = func.__code__

        def _callback(logger, level, msg, *args, **kwargs):
            try:
                self._run_augs(self._create_record(logger, level, msg, args, **kwargs))
            # Don't have unit tests
            except AssertionError:
                raise
            except:  # lgtm[py/catch-base-exception]
                pass

        # This requires a bit of explanation why we're doing this.  Due to how
        # logging itself works we need to pretend that the method actually was
        # created within the logging module.  There are a few ways to detect
        # this and we fake all of them: we use the same function globals (the
        # one from the logging module), we create it entirely there which
        # means that also the filename is set correctly.  This fools the
        # detection code in logging and it makes logging itself skip past our
        # code when determining the code location.
        #
        # Because we point the globals to the logging module we now need to
        # refer to our own functions (original and the crumb recording
        # function) through a closure instead of the global scope.
        #
        # We also add a lot of newlines in front of the code so that the
        # code location lines up again in case someone runs inspect.getsource
        # on the function.
        ns = {}
        eval(compile('''%(offset)sif 1:	
        def factory(original, callback):	
            def %(name)s(self, *args, **kwargs):
                callback(self, %(level)d, *args, **kwargs)
                return original(self, *args, **kwargs)
            return %(name)s	
        \n''' % {
            'offset': '\n' * (code.co_firstlineno - 3),
            'name': name,
            'level': _levelNames[name.upper()]
        }, _logging_srcfile, 'exec'), logging.__dict__, ns)

        new_func = ns['factory'](func, _callback)
        new_func.__doc__ = func.__doc__

        # In theory this should already be set correctly, but in some cases
        # it is not.  So override it.
        new_func.__module__ = func.__module__
        new_func.__name__ = func.__name__

        self._originalMethods[name] = getattr(logging.Logger, name)
        setattr(logging.Logger, name, new_func)

    def _removeHooks(self):
        with self._lock:
            if not self._hooked:
                return

            for name in self.METHODS_TO_HOOK:
                setattr(logging.Logger, name, self._originalMethods[name])
                self._originalMethods[name] = None

            self._hooked = False

    # TODO - This is method is quite expensive and builds the full LogRecordNamespace object for every log call
    # TODO - Change into a lazy namespace to only calculate whatever is needed to evaluate the logging criteria
    def _create_record(self, logger, level, msg, args, exc_info=None, extra=None):
        if _srcfile:
            # IronPython doesn't track Python frames, so findCaller raises an
            # exception on some versions of IronPython. We trap it here so that
            # IronPython can use logging.
            try:
                fn, lno, func = self.findCaller()
            except ValueError:
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        else:
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()
        record = logger.makeRecord(logger.name, level, fn, lno, msg, args, exc_info, func, extra)
        return record

    def findCaller(self):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = currentframe()
        #On some versions of IronPython, currentframe() returns None if
        #IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        rv = "(unknown file)", 0, "(unknown function)"
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == _srcfile or filename == _logging_srcfile:
                f = f.f_back
                continue
            rv = (co.co_filename, f.f_lineno, co.co_name)
            break
        return rv

    def _run_augs(self, record):
        extracted = ContainerNamespace({'log_record': LogRecordNamespace(record)})

        with self._lock:
            for aug in six.itervalues(self._augs):
                aug.execute(None, extracted)




from .dumped_object_namespace import DumpedObjectNamespace


class UnknownNamespace(DumpedObjectNamespace):

    def __init__(self, type, attributes={}):
        super(UnknownNamespace, self).__init__(type, "UnknownObject", attributes)



from threading import Lock

class AugRateLimiter(object):
    ENABLE_SPECIAL_METHODS_FOR_TESTS = False # enable this flag in unittests only to expose some methods that are needed for tests
    def __init__(self, quota, window_size):
        self._quota = quota
        self._window_size = window_size
        self._lock = Lock()

        self._windows = {}

    def allow(self, now_ns):
        if self._quota is None:
            return '0'

        self._cleanup(now_ns)

        current_window_key = now_ns // self._window_size * self._window_size
        prev_window_key = current_window_key - self._window_size

        current_window_usage = self._windows.setdefault(current_window_key, 0)
        prev_window_usage = self._windows.get(prev_window_key)

        if prev_window_usage is None:
            if current_window_usage > self._quota:
                return None
        else:
            prev_weight = 1 - (now_ns - current_window_key) / float(self._window_size)
            weighted_usage = (prev_window_usage * prev_weight) + current_window_usage

            if weighted_usage > self._quota:
                return None

        return current_window_key

    def record(self, key, duration_ns):
        if self._quota is None:
            return

        total_duration = self._windows.get(key)  # windows might be cleared while aug is running (unlikely)

        if total_duration is not None:
            self._windows[key] += duration_ns

    def get_windows_for_tests_only(self):
        if not AugRateLimiter.ENABLE_SPECIAL_METHODS_FOR_TESTS:
            raise NotImplementedError('This method should be used only in unit tests!!!!')
        return self._windows

    def _cleanup(self, now):
        # every 5 windows-times, clear windows older than 5 window-times
        locked = False

        try:
            locked = self._lock.acquire(False)

            if locked:
                if len(self._windows) > 10:
                    self._windows = {k: v for k, v in self._windows.items() if k > now - self._window_size * 5}
        finally:
            if locked:
                self._lock.release()



import time


class TokenBucket(object):
    def __init__(self, limit, interval_seconds, do_once_when_exhausted=None):
        self._remaining = limit
        self._initial_value = limit
        self._last_reset = time.time()
        self._interval = interval_seconds
        self._do_once_when_exhausted = do_once_when_exhausted
        self._do_once_when_exhausted_performed = False

    def is_exhausted(self):
        if time.time() - self._last_reset >= self._interval:
            self._last_reset = time.time()
            self._remaining = self._initial_value
            self._do_once_when_exhausted_performed = False
        return self._remaining < 0

    def do_if_available(self, func):
        self._remaining -= 1
        if self.is_exhausted():
            if self._do_once_when_exhausted is not None and self._do_once_when_exhausted_performed is False:
                self._do_once_when_exhausted_performed = True
                self._do_once_when_exhausted()
            return

        func()



from .namespace import Namespace


class ErrorNamespace(Namespace):

    def __init__(self, message, parameters, exception_namespace, traceback_namespace):
        super(ErrorNamespace, self).__init__()
        self.message = message
        self.parameters = parameters
        self.exc = exception_namespace
        self.traceback = traceback_namespace



from .location_file_line import LocationFileLine
from .location_log_handler import LocationLogHandler
from .http_server_handler import HttpServerHandler
from .location_exception_handler import LocationExceptionHandler

__all__ = (LocationFileLine, LocationLogHandler, HttpServerHandler, LocationExceptionHandler)



"""
from https://github.com/quick2wire/quick2wire-python-api/blob/master/quick2wire/eventfd.py
"""

from ctypes import *  # lgtm[py/polluting-import]
from . import syscall
import os
import errno

# From sys/eventfd.h

EFD_SEMAPHORE = 1
EFD_CLOEXEC = 0o2000000
EFD_NONBLOCK = 0o4000

_libc = CDLL(None, use_errno=True)

eventfd_t = c_uint64

eventfd = syscall.lookup(c_int, "eventfd", (c_uint, c_int))


class Semaphore(syscall.SelfClosing):
    """A Semaphore implemented with eventfd that can be added to a Selector."""

    def __init__(self, count=0, blocking=True):
        """Creates a Semaphore with an initial count.

        Arguments:
        count -- the initial count.
        blocking -- if False calls to wait() do not block if the Semaphore
                    has a count of zero. (default = True)
        """
        self._initial_count = count
        self._flags = EFD_SEMAPHORE | ((not blocking) * EFD_NONBLOCK)
        self._fd = None

    def close(self):
        """Closes the Semaphore and releases its file descriptor."""
        if self._fd is not None:
            os.close(self._fd)
            self._fd = None

    def fileno(self):
        """Returns the Semaphore's file descriptor."""
        if self._fd is None:
            self._fd = eventfd(self._initial_count, self._flags)
        return self._fd

    def signal(self):
        """Signal the semaphore.

        Signalling a semaphore increments its count by one and wakes a
        blocked task that is waiting on the semaphore.
        """
        return os.write(self.fileno(), eventfd_t(1))

    def wait(self):
        """Receive a signal from the Semaphore, decrementing its count by one.

        If the Semaphore is already has a count of zero, either wait
        for a signal if the Semaphore is in blocking mode, or return
        False immediately.

        Returns:
        True  -- the Semaphore received a signal.
        False -- the Semaphore did not receive a signal and is in
                 non-blocking mode.
        """
        try:
            os.read(self.fileno(), 8)
            return True
        except OSError as e:
            if e.errno == errno.EAGAIN:
                return False
            else:
                raise



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: agent_info.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='agent_info.proto',
  package='com.rookout',
  syntax='proto3',
  serialized_options=b'\n\031com.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2',
  serialized_pb=b'\n\x10\x61gent_info.proto\x12\x0b\x63om.rookout\"\xa6\x01\n\x0eSCMInformation\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x0e\n\x06origin\x18\x02 \x01(\t\x12\x37\n\x07sources\x18\x04 \x03(\x0b\x32&.com.rookout.SCMInformation.SourceInfo\x1a\x35\n\nSourceInfo\x12\x17\n\x0fremoteOriginUrl\x18\x01 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x02 \x01(\tJ\x04\x08\x03\x10\x04\"K\n\x12NetworkInformation\x12\x0f\n\x07ip_addr\x18\x01 \x01(\t\x12\x0f\n\x07network\x18\x02 \x01(\t\x12\x13\n\x0b\x65xternal_ip\x18\x03 \x01(\t\"c\n\x11SystemInformation\x12\x10\n\x08hostname\x18\x01 \x01(\t\x12\n\n\x02os\x18\x02 \x01(\t\x12\x12\n\nos_version\x18\x03 \x01(\t\x12\x0e\n\x06\x64istro\x18\x04 \x01(\t\x12\x0c\n\x04\x61rch\x18\x05 \x01(\t\"5\n\x12VersionInformation\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x02 \x01(\t\"I\n\x13PlatformInformation\x12\x10\n\x08platform\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x0f\n\x07variant\x18\x03 \x01(\t\"\xd1\x03\n\x10\x41gentInformation\x12\x10\n\x08\x61gent_id\x18\x01 \x01(\t\x12\x30\n\x07version\x18\x02 \x01(\x0b\x32\x1f.com.rookout.VersionInformation\x12\x30\n\x07network\x18\x03 \x01(\x0b\x32\x1f.com.rookout.NetworkInformation\x12.\n\x06system\x18\x04 \x01(\x0b\x32\x1e.com.rookout.SystemInformation\x12\x32\n\x08platform\x18\x05 \x01(\x0b\x32 .com.rookout.PlatformInformation\x12\x12\n\nexecutable\x18\x06 \x01(\t\x12\x19\n\x11\x63ommand_arguments\x18\x07 \x03(\t\x12\x12\n\nprocess_id\x18\x08 \x01(\r\x12\x39\n\x06labels\x18\t \x03(\x0b\x32).com.rookout.AgentInformation.LabelsEntry\x12(\n\x03scm\x18\n \x01(\x0b\x32\x1b.com.rookout.SCMInformation\x12\x0c\n\x04tags\x18\x0b \x03(\t\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"F\n\x15\x41gentInformationArray\x12-\n\x06\x61gents\x18\x01 \x03(\x0b\x32\x1d.com.rookout.AgentInformationBO\n\x19\x63om.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2b\x06proto3'
)




_SCMINFORMATION_SOURCEINFO = _descriptor.Descriptor(
  name='SourceInfo',
  full_name='com.rookout.SCMInformation.SourceInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='remoteOriginUrl', full_name='com.rookout.SCMInformation.SourceInfo.remoteOriginUrl', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commit', full_name='com.rookout.SCMInformation.SourceInfo.commit', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=141,
  serialized_end=194,
)

_SCMINFORMATION = _descriptor.Descriptor(
  name='SCMInformation',
  full_name='com.rookout.SCMInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='com.rookout.SCMInformation.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='origin', full_name='com.rookout.SCMInformation.origin', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sources', full_name='com.rookout.SCMInformation.sources', index=2,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SCMINFORMATION_SOURCEINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=200,
)


_NETWORKINFORMATION = _descriptor.Descriptor(
  name='NetworkInformation',
  full_name='com.rookout.NetworkInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip_addr', full_name='com.rookout.NetworkInformation.ip_addr', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='network', full_name='com.rookout.NetworkInformation.network', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='external_ip', full_name='com.rookout.NetworkInformation.external_ip', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=202,
  serialized_end=277,
)


_SYSTEMINFORMATION = _descriptor.Descriptor(
  name='SystemInformation',
  full_name='com.rookout.SystemInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hostname', full_name='com.rookout.SystemInformation.hostname', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='os', full_name='com.rookout.SystemInformation.os', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='os_version', full_name='com.rookout.SystemInformation.os_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='distro', full_name='com.rookout.SystemInformation.distro', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='arch', full_name='com.rookout.SystemInformation.arch', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=279,
  serialized_end=378,
)


_VERSIONINFORMATION = _descriptor.Descriptor(
  name='VersionInformation',
  full_name='com.rookout.VersionInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='com.rookout.VersionInformation.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commit', full_name='com.rookout.VersionInformation.commit', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=380,
  serialized_end=433,
)


_PLATFORMINFORMATION = _descriptor.Descriptor(
  name='PlatformInformation',
  full_name='com.rookout.PlatformInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='platform', full_name='com.rookout.PlatformInformation.platform', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='com.rookout.PlatformInformation.version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='variant', full_name='com.rookout.PlatformInformation.variant', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=435,
  serialized_end=508,
)


_AGENTINFORMATION_LABELSENTRY = _descriptor.Descriptor(
  name='LabelsEntry',
  full_name='com.rookout.AgentInformation.LabelsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='com.rookout.AgentInformation.LabelsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.rookout.AgentInformation.LabelsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=931,
  serialized_end=976,
)

_AGENTINFORMATION = _descriptor.Descriptor(
  name='AgentInformation',
  full_name='com.rookout.AgentInformation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent_id', full_name='com.rookout.AgentInformation.agent_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='com.rookout.AgentInformation.version', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='network', full_name='com.rookout.AgentInformation.network', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='system', full_name='com.rookout.AgentInformation.system', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='platform', full_name='com.rookout.AgentInformation.platform', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='executable', full_name='com.rookout.AgentInformation.executable', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='command_arguments', full_name='com.rookout.AgentInformation.command_arguments', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='process_id', full_name='com.rookout.AgentInformation.process_id', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='labels', full_name='com.rookout.AgentInformation.labels', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scm', full_name='com.rookout.AgentInformation.scm', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tags', full_name='com.rookout.AgentInformation.tags', index=10,
      number=11, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_AGENTINFORMATION_LABELSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=511,
  serialized_end=976,
)


_AGENTINFORMATIONARRAY = _descriptor.Descriptor(
  name='AgentInformationArray',
  full_name='com.rookout.AgentInformationArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agents', full_name='com.rookout.AgentInformationArray.agents', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=978,
  serialized_end=1048,
)

_SCMINFORMATION_SOURCEINFO.containing_type = _SCMINFORMATION
_SCMINFORMATION.fields_by_name['sources'].message_type = _SCMINFORMATION_SOURCEINFO
_AGENTINFORMATION_LABELSENTRY.containing_type = _AGENTINFORMATION
_AGENTINFORMATION.fields_by_name['version'].message_type = _VERSIONINFORMATION
_AGENTINFORMATION.fields_by_name['network'].message_type = _NETWORKINFORMATION
_AGENTINFORMATION.fields_by_name['system'].message_type = _SYSTEMINFORMATION
_AGENTINFORMATION.fields_by_name['platform'].message_type = _PLATFORMINFORMATION
_AGENTINFORMATION.fields_by_name['labels'].message_type = _AGENTINFORMATION_LABELSENTRY
_AGENTINFORMATION.fields_by_name['scm'].message_type = _SCMINFORMATION
_AGENTINFORMATIONARRAY.fields_by_name['agents'].message_type = _AGENTINFORMATION
DESCRIPTOR.message_types_by_name['SCMInformation'] = _SCMINFORMATION
DESCRIPTOR.message_types_by_name['NetworkInformation'] = _NETWORKINFORMATION
DESCRIPTOR.message_types_by_name['SystemInformation'] = _SYSTEMINFORMATION
DESCRIPTOR.message_types_by_name['VersionInformation'] = _VERSIONINFORMATION
DESCRIPTOR.message_types_by_name['PlatformInformation'] = _PLATFORMINFORMATION
DESCRIPTOR.message_types_by_name['AgentInformation'] = _AGENTINFORMATION
DESCRIPTOR.message_types_by_name['AgentInformationArray'] = _AGENTINFORMATIONARRAY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SCMInformation = _reflection.GeneratedProtocolMessageType('SCMInformation', (_message.Message,), {

  'SourceInfo' : _reflection.GeneratedProtocolMessageType('SourceInfo', (_message.Message,), {
    'DESCRIPTOR' : _SCMINFORMATION_SOURCEINFO,
    '__module__' : 'agent_info_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.SCMInformation.SourceInfo)
    })
  ,
  'DESCRIPTOR' : _SCMINFORMATION,
  '__module__' : 'agent_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.SCMInformation)
  })
_sym_db.RegisterMessage(SCMInformation)
_sym_db.RegisterMessage(SCMInformation.SourceInfo)

NetworkInformation = _reflection.GeneratedProtocolMessageType('NetworkInformation', (_message.Message,), {
  'DESCRIPTOR' : _NETWORKINFORMATION,
  '__module__' : 'agent_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.NetworkInformation)
  })
_sym_db.RegisterMessage(NetworkInformation)

SystemInformation = _reflection.GeneratedProtocolMessageType('SystemInformation', (_message.Message,), {
  'DESCRIPTOR' : _SYSTEMINFORMATION,
  '__module__' : 'agent_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.SystemInformation)
  })
_sym_db.RegisterMessage(SystemInformation)

VersionInformation = _reflection.GeneratedProtocolMessageType('VersionInformation', (_message.Message,), {
  'DESCRIPTOR' : _VERSIONINFORMATION,
  '__module__' : 'agent_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.VersionInformation)
  })
_sym_db.RegisterMessage(VersionInformation)

PlatformInformation = _reflection.GeneratedProtocolMessageType('PlatformInformation', (_message.Message,), {
  'DESCRIPTOR' : _PLATFORMINFORMATION,
  '__module__' : 'agent_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.PlatformInformation)
  })
_sym_db.RegisterMessage(PlatformInformation)

AgentInformation = _reflection.GeneratedProtocolMessageType('AgentInformation', (_message.Message,), {

  'LabelsEntry' : _reflection.GeneratedProtocolMessageType('LabelsEntry', (_message.Message,), {
    'DESCRIPTOR' : _AGENTINFORMATION_LABELSENTRY,
    '__module__' : 'agent_info_pb2'
    # @@protoc_insertion_point(class_scope:com.rookout.AgentInformation.LabelsEntry)
    })
  ,
  'DESCRIPTOR' : _AGENTINFORMATION,
  '__module__' : 'agent_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AgentInformation)
  })
_sym_db.RegisterMessage(AgentInformation)
_sym_db.RegisterMessage(AgentInformation.LabelsEntry)

AgentInformationArray = _reflection.GeneratedProtocolMessageType('AgentInformationArray', (_message.Message,), {
  'DESCRIPTOR' : _AGENTINFORMATIONARRAY,
  '__module__' : 'agent_info_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.AgentInformationArray)
  })
_sym_db.RegisterMessage(AgentInformationArray)


DESCRIPTOR._options = None
_AGENTINFORMATION_LABELSENTRY._options = None
# @@protoc_insertion_point(module_scope)



from rook.services.logging_location_service import LoggingLocationService

from .location import Location


class LocationLogHandler(Location):

    NAME = 'log_handler'

    def __init__(self, output, aug, arguments, processor_factory):
        super(LocationLogHandler, self).__init__(output, aug)

    def _add_aug_internal(self, trigger_services):
        trigger_services.get_service(LoggingLocationService.NAME).add_logging_aug(self)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Finds all the code objects defined by a module."""

import gc
import os
import sys
import types

import six

# Maximum traversal depth when looking for all the code objects referenced by
# a module or another code object.
_MAX_REFERENTS_BFS_DEPTH = 15

# Absolute limit on the amount of objects to scan when looking for all the code
# objects implemented in a module.
_MAX_VISIT_OBJECTS = 100000

# Maximum referents an object can have before it is skipped in the BFS
# traversal. This is to prevent things like long objects or dictionaries that
# probably do not contain code objects from using the _MAX_VISIT_OBJECTS quota.
_MAX_OBJECT_REFERENTS = 1000

# Object types to ignore when looking for the code objects.
_BFS_IGNORE_TYPES = (types.ModuleType, type(None), bool, float, six.binary_type,
                     six.text_type, types.BuiltinFunctionType,
                     types.BuiltinMethodType, list) + six.integer_types


def GetCodeObjectAtLine(module, line):
  """Searches for a code object at the specified line in the specified module.

  Args:
    module: module to explore.
    line: 1-based line number of the statement.

  Returns:
    (True, Code object) on success or (False, (prev_line, next_line)) on
    failure, where prev_line and next_line are the closest lines with code above
    and below the specified line, or None if they do not exist.
  """
  if not hasattr(module, '__file__'):
    return (False, (None, None))

  prev_line = 0
  next_line = six.MAXSIZE

  for code_object in _GetModuleCodeObjects(module):
    lines = sorted(list(_GetLineNumbers(code_object)))

    for co_line_number in lines:
      if co_line_number == line:
        return (True, code_object)
      elif co_line_number < line:
        prev_line = max(prev_line, co_line_number)
      elif co_line_number > line:
        next_line = min(next_line, co_line_number)
        break

  prev_line = None if prev_line == 0 else prev_line
  next_line = None if next_line == six.MAXSIZE else next_line
  return (False, (prev_line, next_line))


def _GetLineNumbers(code_object):
  """Generator for getting the line numbers of a code object.

  Args:
    code_object: the code object.

  Yields:
    The next line number in the code object.
  """
  # Get the line number deltas, which are the odd number entries, from the
  # lnotab. See
  # https://svn.python.org/projects/python/branches/pep-0384/Objects/lnotab_notes.txt
  # In Python 3, this is just a byte array. In Python 2 it is a string so the
  # numerical values have to be extracted from the individual characters.
  if six.PY3:
    line_incrs = code_object.co_lnotab[1::2]
  else:
    line_incrs = (ord(c) for c in code_object.co_lnotab[1::2])

  current_line = code_object.co_firstlineno

  for line_incr in line_incrs:
    if sys.version_info[:3] >= (3, 6, 0):
      if line_incr >= 0x80:
        line_incr -= 0x100

    current_line += line_incr
    yield current_line


def _GetModuleCodeObjects(module):
  """Gets all code objects defined in the specified module.

  There are two BFS traversals involved. One in this function and the other in
  _FindCodeObjectsReferents. Only the BFS in _FindCodeObjectsReferents has
  a depth limit. This function does not. The motivation is that this function
  explores code object of the module and they can have any arbitrary nesting
  level. _FindCodeObjectsReferents, on the other hand, traverses through class
  definitions and random references. It's much more expensive and will likely
  go into unrelated objects.

  There is also a limit on how many total objects are going to be traversed in
  all. This limit makes sure that if something goes wrong, the lookup doesn't
  hang.

  Args:
    module: module to explore.

  Returns:
    Set of code objects defined in module.
  """

  visit_recorder = _VisitRecorder()
  current = [module]
  code_objects = set()
  while current:
    current = _FindCodeObjectsReferents(module, current, visit_recorder)
    code_objects |= current

    # Unfortunately Python code objects don't implement tp_traverse, so this
    # type can't be used with gc.get_referents. The workaround is to get the
    # relevant objects explicitly here.
    current = [code_object.co_consts for code_object in current]

  return code_objects


def _FindCodeObjectsReferents(module, start_objects, visit_recorder):
  """Looks for all the code objects referenced by objects in start_objects.

  The traversal implemented by this function is a shallow one. In other words
  if the reference chain is a -> b -> co1 -> c -> co2, this function will
  return [co1] only.

  The traversal is implemented with BFS. The maximum depth is limited to avoid
  touching all the objects in the process. Each object is only visited once
  using visit_recorder.

  Args:
    module: module in which we are looking for code objects.
    start_objects: initial set of objects for the BFS traversal.
    visit_recorder: instance of _VisitRecorder class to ensure each object is
        visited at most once.

  Returns:
    List of code objects.
  """
  def CheckIgnoreCodeObject(code_object):
    """Checks if the code object can be ignored.

    Code objects that are not implemented in the module, or are from a lambda or
    generator expression can be ignored.

    If the module was precompiled, the code object may point to .py file, while
    the module says that it originated from .pyc file. We just strip extension
    altogether to work around it.

    Args:
      code_object: code object that we want to check against module.

    Returns:
      True if the code object can be ignored, False otherwise.
    """
    if not hasattr(module, '__file__'):
      return False

    if code_object.co_name in ('<lambda>', '<genexpr>'):
      return True

    code_object_file = os.path.splitext(code_object.co_filename)[0]
    module_file, module_file_ext = os.path.splitext(module.__file__)

    # The simple case.
    if code_object_file == module_file:
      return False

    # If file has been pre-compiled, we only compare base names instead of full paths
    if ".pyc" == module_file_ext and \
            os.path.basename(code_object_file) == os.path.basename(module_file):
      return False

    return True

  def CheckIgnoreClass(cls):
    """Returns True if the class is definitely not coming from "module"."""
    if not hasattr(module, '__file__'):
      return False

    cls_module = sys.modules.get(cls.__module__)
    if not cls_module:
      return False  # We can't tell for sure, so explore this class.

    return (
        cls_module is not module and
        getattr(cls_module, '__file__', None) != module.__file__)

  code_objects = set()
  current = start_objects
  for obj in current:
    visit_recorder.Record(obj)

  depth = 0
  while current and depth < _MAX_REFERENTS_BFS_DEPTH:
    new_current = []
    for current_obj in current:
      referents = gc.get_referents(current_obj)
      if (current_obj is not module.__dict__ and
          len(referents) > _MAX_OBJECT_REFERENTS):
        continue

      for obj in referents:
        try:
          if isinstance(obj, _BFS_IGNORE_TYPES) or not visit_recorder.Record(obj):
            continue

          if isinstance(obj, types.CodeType) and CheckIgnoreCodeObject(obj):
            continue

          if isinstance(obj, six.class_types) and CheckIgnoreClass(obj):
            continue

          if isinstance(obj, types.CodeType):
            code_objects.add(obj)
          else:
            new_current.append(obj)

        except ReferenceError:  # As we are using weak references to iterate, they can disappear
          pass

    current = new_current
    depth += 1

  return code_objects


class _VisitRecorder(object):
  """Helper class to track of already visited objects and implement quota.

  This class keeps a map from integer to object. The key is a unique object
  ID (raw object pointer). The value is the object itself. We need to keep the
  object in the map, so that it doesn't get released during iteration (since
  object ID is only unique as long as the object is alive).
  """

  def __init__(self):
    self._visit_recorder_objects = {}

  def Record(self, obj):
    """Records the object as visited.

    Args:
      obj: visited object.

    Returns:
      True if the object hasn't been previously visited or False if it has
      already been recorded or the quota has been exhausted.
    """
    if len(self._visit_recorder_objects) >= _MAX_VISIT_OBJECTS:
      return False

    # id may potentially invoke end user code, so we protect it with a try/except
    try:
      obj_id = id(obj)
    except Exception:
      return False

    if obj_id in self._visit_recorder_objects:
      return False

    self._visit_recorder_objects[obj_id] = obj
    return True
